'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from .. import spherical as ops
from ... import SphericalGrid, CartesianGrid
from ....fields import ScalarField, VectorField, Tensor2Field



def test_findiff_sph():
    """ test operator for a simple spherical grid """
    grid = SphericalGrid(1.5, 3)
    _, r1, r2 = grid.axes_coords[0]
    assert grid.discretization == (0.5,)
    s = ScalarField(grid, [1, 2, 4]) 
    v = VectorField(grid, [[1, 2, 4], [0]*3, [0]*3]) 

    # test gradient        
    grad = s.gradient(bc='value')
    np.testing.assert_allclose(grad.data[0, :], [1, 3, -6])
    grad = s.gradient(bc='derivative')
    np.testing.assert_allclose(grad.data[0, :], [1, 3, 2])

    # test divergence        
    div = v.divergence(bc='value')
    np.testing.assert_allclose(div.data, [9, 3 + 4/r1, -6 + 8/r2])
    div = v.divergence(bc='derivative')
    np.testing.assert_allclose(div.data, [9, 3 + 4/r1, 2 + 8/r2])
    
    
    
def test_conservative_laplace_sph():
    """ test and compare the two implementation of the laplace operator """
    r_max = 3.14
    for r_min in [0, 0.1]:
        grid = SphericalGrid((r_min, r_max), 8)
        f = ScalarField.from_expression(grid, 'cos(r)')
    
        bcs = grid.get_boundary_conditions('natural')
        lap1 = ops.make_laplace(bcs, conservative=True)
        lap2 = ops.make_laplace(bcs, conservative=False)
        
        np.testing.assert_allclose(lap1(f.data), lap2(f.data),
                                   rtol=0.5, atol=0.5)
        np.testing.assert_allclose(f.apply(lap1).integral, 0, atol=1e-12)
    


@pytest.mark.parametrize('make_op,field,rank',
                         [(ops.make_laplace, ScalarField, 0),
                          (ops.make_divergence, VectorField, 0),
                          (ops.make_gradient, ScalarField, 0),
                          (ops.make_tensor_divergence, Tensor2Field, 1)])
def test_small_annulus(make_op, field, rank):
    """ test whether a small annulus gives the same result as a sphere """
    grids = [SphericalGrid((0, 1), 8),
             SphericalGrid((1e-8, 1), 8),
             SphericalGrid((0.1, 1), 8)]
    
    f = field.random_uniform(grids[0])
    
    res = [make_op(g.get_boundary_conditions(rank=rank))(f.data)
           for g in grids]
    
    np.testing.assert_almost_equal(res[0], res[1], decimal=5)
    assert np.linalg.norm(res[0] - res[2]) > 1e-3




def test_grid_laplace():
    """ test the polar implementation of the laplace operator """
    grid_sph = SphericalGrid(9, 11)
    grid_cart = CartesianGrid([[-5, 5], [-5, 5], [-5, 5]], [12, 10, 11]) 
     
    a_1d = ScalarField.from_expression(grid_sph, 'cos(r)')
    a_3d = a_1d.interpolate_to_grid(grid_cart)

    b_3d = a_3d.laplace('natural')
    b_1d = a_1d.laplace('natural')
    b_1d_3 = b_1d.interpolate_to_grid(grid_cart)
     
    i = slice(1, -1)  # do not compare boundary points
    np.testing.assert_allclose(b_1d_3.data[i, i, i], b_3d.data[i, i, i],
                               rtol=0.2, atol=0.2)
     
     

def test_grid_div_grad():
    """ compare div grad to laplacian """
    grid = SphericalGrid(2*np.pi, 16)
    r = grid.axes_coords[0]
    arr = np.cos(r)

    laplace = grid.get_operator('laplace', 'no-flux')
    grad = grid.get_operator('gradient', 'no-flux')
    div = grid.get_operator('divergence', 'value')
    a = laplace(arr)
    b = div(grad(arr))
    res = -2 * np.sin(r) / r - np.cos(r) 
     
    # do not test the radial boundary points
    np.testing.assert_allclose(a[1:-1], res[1:-1], rtol=0.1, atol=0.1)
    np.testing.assert_allclose(b[1:-1], res[1:-1], rtol=0.1, atol=0.1)       



def test_poisson_solver_spherical():
    """ test the poisson solver on Polar grids """
    grid = SphericalGrid(4, 8)
    for bc_val in ['natural', {'value': 1}]:
        bcs = grid.get_boundary_conditions(bc_val)
        poisson = grid.get_operator('poisson_solver', bcs)
        laplace = grid.get_operator('laplace', bcs)
        
        d = np.random.random(grid.shape)
        d -= ScalarField(grid, d).average  # balance the right hand side
        np.testing.assert_allclose(laplace(poisson(d)), d,
                                   err_msg=f'bcs = {bc_val}')


    grid = SphericalGrid([2, 4], 8)
    for bc_val in ['natural', {'value': 1}]:
        bcs = grid.get_boundary_conditions(bc_val)
        poisson = grid.get_operator('poisson_solver', bcs)
        laplace = grid.get_operator('laplace', bcs)
        
        d = np.random.random(grid.shape)
        d -= ScalarField(grid, d).average  # balance the right hand side
        np.testing.assert_allclose(laplace(poisson(d)), d,
                                   err_msg=f'bcs = {bc_val}')



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
  name = 'py-pde',
  package_data={"pde": ["py.typed"]},
  packages = find_packages(),
  zip_safe=False,  # this is required for mypy to find the py.typed file
  version = '0.8.1',
  license='MIT',
  description = 'Python package for solving partial differential equations',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'David Zwicker',
  author_email = 'david.zwicker@ds.mpg.de',
  url = 'https://github.com/zwicker-group/py-pde',
  download_url = 'https://github.com/zwicker-group/py-pde/archive/v0.8.1.tar.gz',
  keywords = ['pdes', 'partial-differential-equations', 'dynamical-systems'],
  python_requires='>=3.6',
  install_requires=['matplotlib',
                    'numpy',
                    'numba',
                    'scipy',
                    'sympy'],
  extras_require={
        "hdf":  ["h5py>=2"],
        "progress": ["tqdm>=4.40"],
        "interactive": ["napari>=0.3"]
    },
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import functools

import numpy as np

from .. import MemoryStorage, FileStorage
from ...grids import UnitGrid
from ...fields import ScalarField
from ...pdes import DiffusionPDE
from ...tools.misc import module_available



def test_storage_write(tmp_path):
    """ test simple memory storage """
    dim = 5
    grid = UnitGrid([dim])
    field = ScalarField(grid)

    file = tmp_path / "test_storage_write.hdf5"
    
    storage_classes = {'MemoryStorage': MemoryStorage}
    if module_available("h5py"):
        storage_classes['FileStorage'] = functools.partial(FileStorage, file)
    
    for name, storage_cls in storage_classes.items():
        storage = storage_cls(info={'a': 1})
        storage.start_writing(field, info={'b': 2})
        storage.append(np.arange(dim), 0)
        storage.append(np.arange(dim), 1)
        storage.end_writing()
            
        np.testing.assert_allclose(storage.times, np.arange(2))
        for f in storage:
            np.testing.assert_array_equal(f.data, np.arange(dim))
        for i in range(2):
            np.testing.assert_array_equal(storage[i].data, np.arange(dim))
        assert {'a': 1, 'b': 2}.items() <= storage.info.items()
        
        storage = storage_cls()
        storage.clear()
        for i in range(3):
            storage.start_writing(field)
            storage.append(np.arange(dim) + i, i)
            storage.end_writing()
         
        np.testing.assert_allclose(storage.times, np.arange(3),
                                   err_msg='storage class: ' + name)
        
        

def test_storage_truncation(tmp_path):
    """ test whether simple trackers can be used """
    file = tmp_path / "test_storage_truncation.hdf5"
    for truncate in [True, False]:
        storages = [MemoryStorage()]
        if module_available("h5py"):
            storages.append(FileStorage(file))
        tracker_list = [s.tracker(interval=0.01) for s in storages]
      
        grid = UnitGrid([8, 8])
        state = ScalarField.random_uniform(grid, 0.2, 0.3)
        pde = DiffusionPDE()
          
        pde.solve(state, t_range=0.1, dt=0.001, tracker=tracker_list)
        if truncate:
            for storage in storages:
                storage.clear()
        pde.solve(state, t_range=[0.1, 0.2], dt=0.001, tracker=tracker_list)
      
        times = np.arange(0.1, 0.201, 0.01)
        if not truncate:
            times = np.r_[np.arange(0, 0.101, 0.01), times]
        for storage in storages:
            msg = f'truncate={truncate}, storage={storage}'
            np.testing.assert_allclose(storage.times, times, err_msg=msg)
             



'''
.. codeauthor:: David Zwicker <dzwicker@seas.harvard.edu>
'''

from __future__ import division

import copy
import sys
import collections

import pytest
import numpy as np

from .. import cache



def deep_getsizeof(obj, ids=None):
    """Find the memory footprint of a Python object
 
    This is a recursive function that drills down a Python object graph
    like a dictionary holding nested dictionaries with lists of lists
    and tuples and sets.
 
    The sys.getsizeof function does a shallow size of only. It counts each
    object inside a container as pointer only regardless of how big it
    really is.
 
    Function modified from
    https://code.tutsplus.com/tutorials/understand-how-much-memory-your-python-objects-use--cms-25609
    """
    if ids is not None:
        if id(obj) in ids:
            return 0
    else:
        ids = set()
 
    r = sys.getsizeof(obj)
    ids.add(id(obj))
 
    if isinstance(obj, str):
        # simple string
        return r
 
    if isinstance(obj, collections.abc.Mapping):
        # simple mapping
        return r + sum(deep_getsizeof(k, ids) + deep_getsizeof(v, ids)
                       for k, v in obj.items())
 
    if isinstance(obj, collections.abc.Container):
        # collection that is neither a string nor a mapping
        return r + sum(deep_getsizeof(x, ids) for x in obj)
    
    if hasattr(obj, '__dict__'):
        # custom object
        return r + deep_getsizeof(obj.__dict__, ids)
 
    # basic object: neither of the above
    return r 



def get_serialization_methods(with_none=True):
    """ returns possible methods for serialization that are supported """
    methods = ['json', 'pickle']
    
    if with_none:
        methods.append(None)
    
    # check whether yaml is actually available
    try:
        import yaml  # @UnusedImport
    except ImportError:
        pass
    else:
        methods.append('yaml')
        
    return methods



def test_hashes():
    """ test whether the hash key makes sense """
    class Dummy():
        def __init__(self, value):
            self.value = value
            
        def __hash__(self):
            return self.value

    for f in (cache.hash_mutable, cache.hash_readable):
        # test simple objects
        for obj in (1, 1.2, 'a', (1, 2), [1, 2], {1, 2}, {1: 2},
                    {(1, 2): [2, 3], (1, 3): [1, 2]},
                    Dummy(1), np.arange(5)):
            o2 = copy.deepcopy(obj)
            assert f(obj) == f(o2)

        # make sure different objects get different hash
        assert f(1) != f('1')
        assert f('a') != f('b')
        assert f({1, 2}) != f((1, 2))



def test_serializer_nonsense():
    """ test whether errors are thrown for wrong input """
    with pytest.raises(ValueError):
        cache.make_serializer('non-sense')
    with pytest.raises(ValueError):
        cache.make_unserializer('non-sense')



@pytest.mark.parametrize('method', get_serialization_methods())
def test_serializer(method):
    """ tests whether the make_serializer returns a canonical hash """
    encode = cache.make_serializer(method)
    
    assert encode(1) == encode(1)
    
    assert encode([1, 2, 3]) != encode([2, 3, 1])
    if method != 'json':
        # json cannot encode sets
        assert encode({1, 2, 3}) == encode({2, 3, 1})



def test_serializer_hash_mutable():
    """ tests whether the make_serializer returns a canonical hash """
    # test special serializer
    encode = cache.make_serializer('hash_mutable')
    assert encode({'a': 1, 'b': 2}) == encode({'b': 2, 'a': 1})
    
    c1 = collections.OrderedDict([('a', 1), ('b', 2)])
    c2 = collections.OrderedDict([('b', 2), ('a', 1)])
    assert cache.hash_mutable(c1) != cache.hash_mutable(c2)
    assert cache.hash_mutable(dict(c1)) == cache.hash_mutable(dict(c2))
    
    
    class Test():
        """ test class that neither implements __eq__ nor __hash__ """
        def __init__(self, a):
            self.a = a
            
    assert cache.hash_mutable(Test(1)) != cache.hash_mutable(Test(1)) 

    class TestEq():
        """ test class that only implements __eq__ and not __hash__ """
        def __init__(self, a):
            self.a = a
            
        def __eq__(self, other):
            return self.a == other.a
            
    assert cache.hash_mutable(TestEq(1)) == cache.hash_mutable(TestEq(1)) 
    assert cache.hash_mutable(TestEq(1)) != cache.hash_mutable(TestEq(2)) 



def test_unserializer():
    """ tests whether the make_serializer and make_unserializer return the 
    original objects """
    data_list = [None, 1, [1, 2], {'b': 1, 'a': 2}]
    for method in get_serialization_methods():
        encode = cache.make_serializer(method)
        decode = cache.make_unserializer(method)
        for data in data_list:
            assert data == decode(encode(data))



def _test_SerializedDict(storage, reinitialize=None,
                         key_serialization='pickle',
                         value_serialization='pickle'):
    """ tests the SerializedDict class with a particular parameter set """
    data = cache.SerializedDict(key_serialization, value_serialization,
                                storage_dict=storage)
    
    if value_serialization == 'none':
        with pytest.raises(TypeError):
            data['a'] = 1
            
        v1, v2, v3 = '1', '2', '3'
            
    else:
        v1, v2, v3 = 1, 2, '3'
        
    data['a'] = v1
    assert len(data) == v1
    data['b'] = v2
    assert data['b'] == v2

    assert len(data) == v2
    del data['a']
    assert len(data) == v1
    with pytest.raises(KeyError):
        data['a']
    
    data.update({'d': v3})
    assert len(data) == v2
    
    # reinitialize the storage dictionary
    if reinitialize is not None:
        data._data = reinitialize()
    assert len(data) == v2
    assert data['b'] == v2
    assert 'd' in data
    assert {'b', 'd'} == set(data.keys())
    assert {v2, v3} == set(data.values())
    data.clear()
    assert len(data) == 0
    
    # reinitialize the dictionary
    if reinitialize is not None:
        data._data = reinitialize()
    assert len(data) == 0
    


@pytest.mark.parametrize('cache_storage', [None, "get_finite_dict"])
def test_property_cache(cache_storage):
    """ test cached_property decorator """
    
    # create test class
    class CacheTest():
        """ class for testing caching """
        
        def __init__(self):
            self.counter = 0
        
        def get_finite_dict(self, n):
            return cache.DictFiniteCapacity(capacity=1)
        
        @property
        def uncached(self):
            self.counter += 1
            return 1
        
        def cached(self):
            self.counter += 1
            return 2    
    
    # apply the cache with the given storage
    if cache_storage is None: 
        decorator = cache.cached_property()
    else:
        decorator = cache.cached_property(cache_storage)
    CacheTest.cached = decorator(CacheTest.cached)
        
    # try to objects to make sure caching is done on the instance level
    for obj in [CacheTest(), CacheTest()]:        
        # test uncached method
        assert obj.uncached == 1
        assert obj.counter == 1
        assert obj.uncached == 1
        assert obj.counter == 2
        obj.counter = 0
        
        # test cached methods
        assert obj.cached == 2
        assert obj.counter == 1
        assert obj.cached == 2
        assert obj.counter == 1


        
@pytest.mark.parametrize('serializer',
                         get_serialization_methods(with_none=False)) 
@pytest.mark.parametrize('cache_factory', [None, 'get_finite_dict'])
def test_method_cache(serializer, cache_factory):
    """ test one particular parameter set of the cached_method decorator """
    
    # create test class
    class CacheTest():
        """ class for testing caching """
        
        def __init__(self):
            self.counter = 0
        
        def get_finite_dict(self, name):
            return cache.DictFiniteCapacity(capacity=1)
        
        def uncached(self, arg):
            self.counter += 1
            return arg
        
        @cache.cached_method(hash_function=serializer,
                             factory=cache_factory)
        def cached(self, arg):
            self.counter += 1
            return arg    
        
        @cache.cached_method(hash_function=serializer,
                             factory=cache_factory)
        def cached_kwarg(self, a=0, b=0):
            self.counter += 1
            return a + b
        
    # test what happens when the decorator is applied wrongly
    with pytest.raises(ValueError):
        cache.cached_method(CacheTest.cached)
        
    # try to objects to make sure caching is done on the instance level and
    # that clearing the cache works
    obj1, obj2 = CacheTest(), CacheTest()
    for k, obj in enumerate([obj1, obj2, obj1]):        
        
        # clear the cache before the first and the last pass
        if k == 0 or k == 2:
            CacheTest.cached.clear_cache_of_obj(obj)
            CacheTest.cached_kwarg.clear_cache_of_obj(obj)
            obj.counter = 0
        
        # test uncached method
        assert obj.uncached(1) == 1
        assert obj.counter == 1
        assert obj.uncached(1) == 1
        assert obj.counter == 2
        obj.counter = 0
        
        # test cached methods
        for method in (obj.cached, obj.cached_kwarg):
            # run twice to test clearing the cache
            for _ in (None, None):
                # test simple caching behavior
                assert method(1) == 1
                assert obj.counter == 1
                assert method(1) == 1
                assert obj.counter == 1
                assert method(2) == 2
                assert obj.counter == 2
                assert method(2) == 2
                assert obj.counter == 2
                
                # test special properties of cache_factories
                if cache_factory is None:
                    assert method(1) == 1
                    assert obj.counter == 2
                elif cache_factory == 'get_finite_dict':
                    assert method(1) == 1
                    assert obj.counter == 3
                else:
                    raise ValueError('Unknown cache_factory `%s`'
                                     % cache_factory)
    
                obj.counter = 0
                # clear cache to test the second run
                method.clear_cache_of_obj(obj)
        
        # test complex cached method
        assert obj.cached_kwarg(1, b=2) == 3
        assert obj.counter == 1
        assert obj.cached_kwarg(1, b=2) == 3
        assert obj.counter == 1
        assert obj.cached_kwarg(2, b=2) == 4
        assert obj.counter == 2
        assert obj.cached_kwarg(2, b=2) == 4
        assert obj.counter == 2
        assert obj.cached_kwarg(1, b=3) == 4
        assert obj.counter == 3
        assert obj.cached_kwarg(1, b=3) == 4
        assert obj.counter == 3
    


@pytest.mark.parametrize('serializer',
                         get_serialization_methods(with_none=False)) 
@pytest.mark.parametrize('cache_factory', [None, 'get_finite_dict'])
def test_method_cache_extra_args(serializer, cache_factory):
    """ test extra arguments in the cached_method decorator """
    # create test class
    class CacheTest():
        """ class for testing caching """
        
        def __init__(self, value=0):
            self.counter = 0
            self.value = 0
        
        def get_finite_dict(self, name):
            return cache.DictFiniteCapacity(capacity=1)
        
        @cache.cached_method(hash_function=serializer, extra_args=['value'],
                             factory=cache_factory)
        def cached(self, arg):
            self.counter += 1
            return self.value + arg    
        
    obj = CacheTest(0)
        
    # test simple caching behavior
    assert obj.cached(1) == 1
    assert obj.counter == 1
    assert obj.cached(1) == 1
    assert obj.counter == 1
    assert obj.cached(2) == 2
    assert obj.counter == 2
    assert obj.cached(2) == 2
    assert obj.counter == 2

    obj.value = 10
    # test simple caching behavior
    assert obj.cached(1) == 11
    assert obj.counter == 3
    assert obj.cached(1) == 11
    assert obj.counter == 3
    assert obj.cached(2) == 12
    assert obj.counter == 4
    assert obj.cached(2) == 12
    assert obj.counter == 4
     


@pytest.mark.parametrize('serializer',
                         get_serialization_methods(with_none=False)) 
@pytest.mark.parametrize('cache_factory', [None, 'get_finite_dict'])
@pytest.mark.parametrize('ignore_args', ['display', ['display']])
def test_method_cache_ignore(serializer, cache_factory, ignore_args):
    """ test ignored parameters of the cached_method decorator """
    # create test class
    class CacheTest():
        """ class for testing caching """
        
        def __init__(self):
            self.counter = 0
        
        def get_finite_dict(self, name):
            return cache.DictFiniteCapacity(capacity=1)
        
        @cache.cached_method(hash_function=serializer,
                             ignore_args=ignore_args,
                             factory=cache_factory)
        def cached(self, arg, display=True):
            self.counter += 1
            return arg    
        
    obj = CacheTest()
        
    # test simple caching behavior
    assert obj.cached(1, display=True) == 1
    assert obj.counter == 1
    assert obj.cached(1, display=True) == 1
    assert obj.counter == 1
    assert obj.cached(1, display=False) == 1
    assert obj.counter == 1
    assert obj.cached(2, display=True) == 2
    assert obj.counter == 2
    assert obj.cached(2, display=False) == 2
    assert obj.counter == 2
    assert obj.cached(2, display=False) == 2
    assert obj.counter == 2



def test_cache_clearing():
    """ make sure that memory is freed when cache is cleared """
    class Test():
        """ simple test object with a cache """
        @cache.cached_method()
        def calc(self, n):
            return np.empty(n)
        
        def clear_cache(self):
            self._cache_methods = {}
            
        def clear_specific(self):
            self.calc.clear_cache_of_obj(self)
        
    t = Test()
    
    mem0 = deep_getsizeof(t)
    
    for clear_cache in (t.clear_cache, t.clear_specific):
        t.calc(100)
        mem1 = deep_getsizeof(t)
        assert mem1 > mem0
        t.calc(200)
        mem2 = deep_getsizeof(t)
        assert mem2 > mem1
        t.calc(100)
        mem3 = deep_getsizeof(t)
        assert mem3 == mem2
    
        clear_cache()
        mem4 = deep_getsizeof(t)
        assert mem4 >= mem0
        assert mem1 >= mem4



def test_serialized_dict():
    """ test SerializedDict """
    d = cache.SerializedDict()
    assert len(d) == 0
    d['a'] = 1
    assert len(d) == 1
    assert 'a' in d
    assert d['a'] == 1
    assert list(d) == ['a']
    del d['a']
    assert len(d) == 0
    
    
    
def test_finite_dict():
    """ test DictFiniteCapacity """
    d = cache.DictFiniteCapacity(capacity=1)
    d['a'] = 1
    assert d['a'] == 1
    d['b'] = 2
    assert d['b'] == 2
    assert 'a' not in d
    
    d1 = cache.DictFiniteCapacity(capacity=1)
    d2 = cache.DictFiniteCapacity(capacity=1)
    d1['a'] = d2['a'] = 1
    assert d1 == d2
    d2 = cache.DictFiniteCapacity(capacity=2)
    d2['a'] = 1
    assert d1 != d2
    


"""
A simple diffusion equation

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numpy as np


from .base import PDEBase
from ..fields import ScalarField, FieldCollection
from ..grids.boundaries.axes import BoundariesData
from ..tools.numba import nb, jit
from ..tools.docstrings import fill_in_docstring


        
class WavePDE(PDEBase):
    r""" A simple wave equation
    
    The mathematical definition,
    
    .. math::
        \partial_t^2 u = c^2 \nabla^2 u
        
    is implemented as two first-order equations:
    
    .. math::
        \partial_t u &= v \\
        \partial_t v &= c^2 \nabla^2 u
        
        
    where :math:`u` is the density field that and :math:`c` sets the wave speed.
    """

    explicit_time_dependence = False
    

    @fill_in_docstring
    def __init__(self, speed: float = 1,
                 bc: BoundariesData = 'natural'):
        """ 
        Args:
            speed (float):
                The speed :math:`c` of the wave
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
        """
        super().__init__()
        
        self.speed = speed
        self.bc = bc
        
        
    def get_initial_condition(self, u: ScalarField, v: ScalarField = None):
        """ create a suitable initial condition
        
        Args:
            u (:class:`~pde.fields.ScalarField`):
                The initial density on the grid
            v (:class:`~pde.fields.ScalarField`, optional):
                The initial rate of change. This is assumed to be zero if the
                value is omitted.
                
        Returns:
            :class:`~pde.fields.FieldCollection`:
                The combined fields u and v, suitable for the simulation
        """
        if v is None:
            v = u.copy(data=0)
        return FieldCollection([u, v])
            
            
    def evolution_rate(self, state: FieldCollection,  # type: ignore
                       t: float = 0) -> FieldCollection:
        """ evaluate the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldCollection`):
                The fields :math:`u` and :math:`v` distribution
            t (float):
                The current time point
            
        Returns:
            :class:`~pde.fields.FieldCollection`:
            Scalar field describing the evolution rate of the PDE 
        """
        assert isinstance(state, FieldCollection)
        u, v = state
        u_t = v.copy()
        v_t = self.speed**2 * u.laplace('natural')  # type: ignore
        return FieldCollection([u_t, v_t])
    
    
    def _make_pde_rhs_numba(self, state: FieldCollection  # type: ignore
                            ) -> Callable:
        """ create a compiled function evaluating the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldCollection`):
                An example for the state defining the grid and data types
                
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.  
        """
        shape = state.grid.shape
        arr_type = nb.typeof(np.empty((2,) + shape, dtype=np.double))
        signature = arr_type(arr_type, nb.double)
        
        speed2 = self.speed**2
        laplace = state.grid.get_operator('laplace', bc=self.bc)

        @jit(signature)
        def pde_rhs(state_data: np.ndarray, t: float):
            """ compiled helper function evaluating right hand side """
            rate = np.empty_like(state_data)
            rate[0] = state_data[1]
            laplace(state_data[0], out=rate[1])
            rate[1] *= speed2
            return rate
            
        return pde_rhs  # type: ignore
    



'''
Created on Mar 21, 2020

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import textwrap
import re
from typing import TypeVar



DOCSTRING_REPLACEMENTS = {
    # description of function arguments
    'ARG_BOUNDARIES_INSTANCE': """
        Specifies the boundary conditions applied to the field. This must be an
        instance of :class:`~pde.grids.boundaries.axes.Boundaries`, which can be
        created from various data formats using the class method
       :func:`~pde.grids.boundaries.axes.Boundaries.from_data`."""[1:],
       
    'ARG_BOUNDARIES': """
        Boundary conditions are generally given as a list with one condition for
        each axis. For periodic axis, only periodic boundary conditions are
        allowed (indicated by 'periodic'). For non-periodic axes, different 
        boundary conditions can be specified for the lower and upper end (by a
        tuple of two conditions). For instance, Dirichlet conditions enforcing a
        value NUM (specified by `{'value': NUM}`) and Neumann conditions 
        enforcing the value DERIV for the derivative in the normal direction
        (specified by `{'derivative': DERIV}`) are supported.Note that the
        special value 'natural' imposes periodic boundary conditions for
        periodic axis and a vanishing derivative otherwise. More information can
        be found in the
        :ref:`boundaries documentation <documentation-boundaries>`."""[1:],

    'ARG_TRACKER_INTERVAL': """
        Determines how often the tracker interrupts the simulation. Simple
        numbers are interpreted as durations measured in the simulation time
        variable. Alternatively, a string using the format 'hh:mm:ss' can be
        used to give durations in real time. Finally, instances of the classes
        defined in :mod:`~pde.trackers.intervals` can be given for more control.
        """[1:-1],
        
    'ARG_PLOT_QUANTITIES': """
        A 2d list of quantities that are shown in a rectangular arrangement.
        If `quantities` is a simple list, the panels will be rendered as a
        single row.
        Each panel is defined by a dictionary, where the mandatory item 'source'
        defines what is being shown. Here, an integer specifies the component
        that is extracted from the field while a function is evaluate with the
        full state as an input and the result is shown.
        Additional items in the dictionary can be 'title' (setting the title of
        the panel), 'scale' (defining the color range shown; these are typically
        two numbers defining the lower and upper bound, but if only one is given
        the range [0, scale] is assumed), and 'cmap' (defining the colormap
        being used)."""[1:],
        
    'ARG_PLOT_SCALE': """
        Flag determining how the range of the color scale is determined. In the
        simplest case a tuple of numbers marks the lower and upper end of the 
        scalar values that will be shown. If only a single number is supplied,
        the range starts at zero and ends at the given number. Additionally, the
        special value 'automatic' determines the range from the range of scalar
        values."""[1:-1],
    
    # descriptions of the discretization and the symmetries         
    'DESCR_CYLINDRICAL_GRID': r"""
        The cylindrical grid assumes polar symmetry, so that fields only depend
        on the radial coordinate `r` and the axial coordinate `z`. Here, the
        first axis is along the radius, while the second axis is along the axis
        of the cylinder. The radial discretization is defined as
        :math:`r_i = (i + \frac12) \Delta r` for :math:`i=0, \ldots, N_r-1`.
        """[1:-1],
        
    'DESCR_POLAR_GRID': r"""
        The polar grid assumes polar symmetry, so that fields only depend on the
        radial coordinate `r`. The radial discretization is defined as
        :math:`r_i = r_\mathrm{min} + (i + \frac12) \Delta r` for
        :math:`i=0, \ldots, N_r-1`,  where :math:`r_\mathrm{min}` is the radius
        of the inner boundary, which is zero by default. Note that the radius of
        the outer boundary is given by
        :math:`r_\mathrm{max} = r_\mathrm{min} + N_r \Delta r`."""[1:],
        
    'DESCR_SPHERICAL_GRID': r"""
        The spherical grid assumes spherical symmetry, so that fields only
        depend on the radial coordinate `r`. The radial discretization is
        defined as :math:`r_i = r_\mathrm{min} + (i + \frac12) \Delta r` for
        :math:`i=0, \ldots, N_r-1`,  where :math:`r_\mathrm{min}` is the radius
        of the inner boundary, which is zero by default. Note that the radius of
        the outer boundary is given by
        :math:`r_\mathrm{max} = r_\mathrm{min} + N_r \Delta r`."""[1:],
        
    # notes in the docstring
    'WARNING_EXEC': r"""
        This implementation uses :func:`exec` and should therefore not be used 
        in a context where malicious input could occur."""[1:]
}



def get_text_block(identifier: str) -> str:
    """ return a single text block
    
    Args:
        identifier (str): The name of the text block
        
    Returns:
        str: the text block as one long line.
    """
    raw_text = DOCSTRING_REPLACEMENTS[identifier]
    return ''.join(textwrap.dedent(raw_text))



TFunc = TypeVar('TFunc')


def fill_in_docstring(f: TFunc) -> TFunc:
    """ decorator that replaces text in the docstring of a function """
    # initialize textwrapper for formatting docstring
    tw = textwrap.TextWrapper(width=80, expand_tabs=True,
                              replace_whitespace=True, drop_whitespace=True)
            
    docstring = f.__doc__
    for name, value in DOCSTRING_REPLACEMENTS.items():
        
        def repl(matchobj) -> str:
            """ helper function replacing token in docstring """
            tw.initial_indent = tw.subsequent_indent = matchobj.group(1)
            return tw.fill(textwrap.dedent(value))

        # replace the token with the correct indentation
        token = '{' + name + '}'
        docstring = re.sub(f"^([ \t]*){token}", repl, docstring,  # type: ignore
                           flags=re.MULTILINE)

    f.__doc__ = docstring
    return f



'''
Miscallenous python functions 

.. autosummary::
   :nosignatures:

   module_available
   environment
   ensure_directory_exists
   preserve_scalars
   decorator_arguments
   skipUnlessModule
   get_progress_bar_class
   display_progress
   add_scaled_colorbar
   import_class
   classproperty
   hybridmethod
   estimate_computation_speed
   hdf_write_attributes
   in_jupyter_notebook

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import os
import errno
import functools
import json
import sys
import importlib
import unittest
import warnings
from pathlib import Path
from typing import Callable, Dict, Any, List, Union

import numpy as np



def module_available(module_name: str) -> bool:
    """ check whether a python module is available
    
    Args:
        module_name (str): The name of the module
        
    Returns:
        `True` if the module can be imported and `False` otherwise
    """
    try:
        importlib.import_module(module_name)
    except ImportError:
        return False
    else:
        return True



def environment(dict_type=dict) -> Dict[str, Any]:
    """ obtain information about the compute environment
    
    Args:
        dict_type: The type to create the returned dictionaries. The default is
            `dict`, but :class:`collections.OrderedDict` is an alternative.
    
    Returns:
        dict: information about the python installation and packages
    """
    from .. import __version__ as package_version

    from .numba import numba_environment

    def get_package_versions(packages: List[str]) -> Dict[str, str]:
        """ tries to load certain python packages and returns their version """
        versions: Dict[str, str] = dict_type()
        for name in sorted(packages):
            try:
                module = importlib.import_module(name)
            except ImportError:
                versions[name] = 'not available'
            else:
                versions[name] = module.__version__  # type: ignore
        return versions

    result: Dict[str, Any] = dict_type()
    result['package version'] = package_version
    result['python version'] = sys.version
    result['mandatory packages'] = get_package_versions(
                            ['matplotlib', 'numba', 'numpy', 'scipy', 'sympy'])
    result['optional packages'] = get_package_versions(['h5py', 'pandas',
                                                        'pyfftw', 'tqdm'])
    if module_available('numba'):
        result['numba environment'] = numba_environment()
    
    return result
    
    
    

def ensure_directory_exists(folder: Union[str, Path]):
    """ creates a folder if it not already exists
    
    Args:
        folder (str): path of the new folder
    """
    folder = str(folder)
    if folder == '':
        return
    try:
        os.makedirs(folder)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise
    


def preserve_scalars(method: Callable) -> Callable:
    """ decorator that makes vectorized methods work with scalars
    
    This decorator allows to call functions that are written to work on numpy
    arrays to also accept python scalars, like `int` and `float`. Essentially,
    this wrapper turns them into an array and unboxes the result.

    Args:
        method: The method being decorated
        
    Returns:
        The decorated method
    """
    @functools.wraps(method)
    def wrapper(self, *args):
        args = [np.asanyarray(arg, dtype=np.double)
                for arg in args]
        if args[0].ndim == 0:
            args = [arg[None] for arg in args]
            return method(self, *args)[0]
        else:
            return method(self, *args)
    return wrapper



def decorator_arguments(decorator: Callable) -> Callable:
    r""" make a decorator usable with and without arguments:
    
    The resulting decorator can be used like `@decorator`
    or `@decorator(\*args, \**kwargs)`
    
    Inspired by https://stackoverflow.com/a/14412901/932593
    
    Args:
        decorator: the decorator that needs to be modified
        
    Returns:
        the decorated function
    """
    @functools.wraps(decorator)
    def new_decorator(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return decorator(args[0])
        else:
            # decorator arguments
            return lambda realf: decorator(realf, *args, **kwargs)

    return new_decorator



def skipUnlessModule(module_name: str) -> Callable:
    """ decorator that skips a test when a module is not available
    
    Args:
        module_name (str): The name of the required module
        
    Returns:
        A function, so this can be used as a decorator
    """
    if module_available(module_name):
        # return no-op decorator
        def wrapper(f: Callable) -> Callable:
            return f
        return wrapper
    else:
        # return decorator skipping test
        return unittest.skip(f"requires {module_name}")
        


class MockProgress():
    """ indicates progress by printing dots to stderr """
    def __init__(self, iterable=None, *args, **kwargs):
        self.iterable = iterable
        self.n = self.total = 0
        self.disable = False

    def __iter__(self):
        return iter(self.iterable)

    def close(self, *args, **kwargs): pass
    
    def refresh(self, *args, **kwargs): 
        sys.stderr.write('.')
        sys.stderr.flush()
        
    def set_description(self, msg: str, refresh: bool = True, *args,
                        **kwargs):
        if refresh:
            self.refresh()



def get_progress_bar_class():
    """ returns a class that behaves as progress bar.
    
    This either uses classes from the optional `tqdm` package or a simple
    version that writes dots to stderr, if the class it not available.
    """
    try:
        # try importing the tqdm package
        import tqdm
        
    except ImportError:
        # create a mock class, since tqdm is not available 
        # progress bar package does not seem to be available
        warnings.warn('`tqdm` package is not available. Progress will '
                      'be indicated by dots.')
        progress_bar_class = MockProgress

    else:
        # tqdm is available => decide which class to return   
        tqdm_version = tuple(int(v) for v in tqdm.__version__.split('.')[:2])
        if tqdm_version >= (4, 40):
            # optionally import notebook progress bar in recent version
            try:
                # check whether progress bar can use a widget
                import ipywidgets  # @UnusedImport  
            except ImportError:
                # widgets are not available => use standard tqdm
                progress_bar_class = tqdm.tqdm
            else:
                # use the fancier version of the progress bar in jupyter
                from tqdm.auto import tqdm as progress_bar_class
        else:
            # only import text progress bar in older version
            progress_bar_class = tqdm.tqdm
            warnings.warn('Your version of tqdm is outdated. To get a nicer '
                          'progress bar update to at least version 4.40.')

    return progress_bar_class
        
        

def display_progress(iterator, total=None, enabled=True, **kwargs):
    r"""
    displays a progress bar when iterating
    
    Args:
        iterator (iter): The iterator
        total (int): Total number of steps
        enabled (bool): Flag determining whether the progress is display
        **kwargs: All extra arguments are forwarded to the progress bar class
        
    Returns:
        A class that behaves as the original iterator, but shows the progress
        alongside iteration.
    """
    if not enabled:
        return iterator
    
    return get_progress_bar_class()(iterator, total=total, **kwargs)



def add_scaled_colorbar(im,
                        ax=None,
                        aspect: float = 20,
                        pad_fraction: float = 0.5,
                        **kwargs):
    """ add a vertical color bar to an image plot
    
    The height of the colorbar is now adjusted to the plot, so that the width
    determined by `aspect` is now given relative to the height. Moreover, the
    gap between the colorbar and the plot is now given in units of the fraction
    of the width by `pad_fraction`. 

    Inspired by https://stackoverflow.com/a/33505522/932593
    
    Args:
        im: object returned from :meth:`matplotlib.pyplot.imshow`
        ax (:class:`matplotlib.axes.Axes`): the current figure axes
        aspect (float): the target aspect ratio of the colorbar
        pad_fraction (float): Width of the gap between colorbar and image
        **kwargs: Additional parameters are passed to colorbar call

    Returns:
        the result of the colorbar call
    """
    from mpl_toolkits import axes_grid1
    divider = axes_grid1.make_axes_locatable(im.axes)
    width = axes_grid1.axes_size.AxesY(im.axes, aspect=1./aspect)
    pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
    cax = divider.append_axes("right", size=width, pad=pad)
#     if ax is not None:
#         plt.sca(ax)
    return im.axes.figure.colorbar(im, ax=ax, cax=cax, **kwargs)            
       


def import_class(identifier: str):
    """ import a class or module given an identifier 
    
    Args:
        identifier (str):
            The identifier can be a module or a class. For instance, calling the
            function with the string `identifier == 'numpy.linalg.norm'` is
            roughly equivalent to running `from numpy.linalg import norm` and
            would return a reference to `norm`.
    """
    module_path, _, class_name = identifier.rpartition('.')
    if module_path:
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    else:
        # this happens when identifier does not contain a dot
        return importlib.import_module(class_name)
    


class classproperty:
    """ decorator that can be used to define read-only properties for classes. 
    Adopted from http://stackoverflow.com/a/5192374/932593
    """
    def __init__(self, f):
        self.f = f
         
    def __get__(self, obj, owner):
        return self.f(owner)
    


class hybridmethod:
    """
    descriptor that can be used as a decorator to allow calling a method both
    as a classmethod and an instance method
     
    Adapted from https://stackoverflow.com/a/28238047
    """
    
    def __init__(self, fclass, finstance=None, doc=None):
        self.fclass = fclass
        self.finstance = finstance
        self.__doc__ = doc or fclass.__doc__
        # support use on abstract base classes
        self.__isabstractmethod__ = bool(
            getattr(fclass, '__isabstractmethod__', False)
        )
        

    def classmethod(self, fclass):
        return type(self)(fclass, self.finstance, None)


    def instancemethod(self, finstance):
        return type(self)(self.fclass, finstance, self.__doc__)


    def __get__(self, instance, cls):
        if instance is None or self.finstance is None:
            # either bound to the class, or no instance method available
            return self.fclass.__get__(cls, None)
        return self.finstance.__get__(instance, cls)    
    
    
    
def estimate_computation_speed(func: Callable, *args, **kwargs) -> float:
    """ estimates the computation speed of a function
    
    Args:
        func (callable): The function to call
    
    Returns:
        float: the number of times the function can be calculated in one second.
        The inverse is thus the runtime in seconds per function call
    """
    import timeit
    test_duration = kwargs.pop('test_duration', 1)
    
    # prepare the function
    if args or kwargs:
        test_func = functools.partial(func, *args, **kwargs)
    else:
        test_func = func  # type: ignore
    
    # call function once to allow caches be filled
    test_func()
     
    # call the function until the total time is achieved
    number, duration = 1, 0
    while duration < 0.1 * test_duration:
        number *= 10
        duration = timeit.timeit(test_func, number=number)  # type: ignore
    return number / duration



def hdf_write_attributes(hdf_path,
                         attributes: Dict[str, Any] = None,
                         raise_serialization_error: bool = False) -> None:
    """ write (JSON-serialized) attributes to a hdf file
    
    Args:
        hdf_path:
            Path to a group or dataset in an open HDF file
        attributes (dict):
            Dictionary with values written as attributes
        raise_serialization_error (bool):
            Flag indicating whether serialization errors are raised or silently
            ignored
    """ 
    if attributes is None:
        return
        
    for key, value in attributes.items():
        try:
            value_serialized = json.dumps(value)
        except TypeError:
            if raise_serialization_error:
                raise
        else:
            hdf_path.attrs[key] = value_serialized



def in_jupyter_notebook() -> bool:
    """ checks whether we are in a jupyter notebook """
    try:
        from IPython import get_ipython
    except ImportError:
        return False
        
    try:
        ipython_config = get_ipython().config
    except AttributeError:
        return False
        
    return ('IPKernelApp' in ipython_config)



"""
Defines a class storing data on the file system using the hierarchical data
format (hdf).

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

import json
import logging
from pathlib import Path
from typing import Optional, Tuple, Any  # @UnusedImport

import numpy as np

from .base import StorageBase, InfoDict
from ..fields.base import FieldBase
from ..tools.misc import ensure_directory_exists, hdf_write_attributes



class FileStorage(StorageBase):
    """ store data in a hdf5 file """
    
    def __init__(self, filename: str,
                 info: InfoDict = None,
                 write_mode: str = 'truncate_once',
                 max_length: Optional[int] = None,
                 compression: bool = True,
                 keep_opened: bool = True):
        """
        Args:
            filename (str):
                The path to the hdf5-file where the data is stored
            info (dict):
                Supplies extra information that is stored in the storage
            write_mode (str):
                Determines how new data is added to already existing data.
                Possible values are: 'append' (data is always appended),
                'truncate' (data is cleared every time this storage is used
                for writing), or 'truncate_once' (data is cleared for the first
                writing, but appended subsequently). Alternatively, specifying
                'readonly' will disable writing completely.
            max_length (int, optional):
                Maximal number of entries that will be stored in the file. This
                can be used to preallocate data, which can lead to smaller
                files, but is also less flexible. Giving `max_length = None`,
                allows for arbitrarily large data, which might lead to larger
                files.
            compression (bool):
                Whether to store the data in compressed form. Automatically
                enabled chunked storage.
            keep_opened (bool):
                Flag indicating whether the file should be kept opened after
                each writing. If `False`, the file will be closed after writing
                a dataset. This keeps the file in a consistent state, but also
                requires more work before data can be written.
        """
        super().__init__(info=info, write_mode=write_mode)
        self.filename = Path(filename)
        self.compression = compression
        self.keep_opened = keep_opened
        
        self._logger = logging.getLogger(self.__class__.__name__)
        self._file: Any = None
        self._is_writing = False
        self._data_length: int = None  # type: ignore
        self._max_length: Optional[int] = max_length
        
        if self.filename.is_file() and self.filename.stat().st_size > 0:
            try:
                self._open('reading')
            except (OSError, KeyError):
                self.close()
                self._logger.warning(f'File `{filename}` could not be opened '
                                     'for reading')

    
    @property
    def _file_state(self) -> str:
        """ str: the state that the file is currently in """
        if self._file is None:
            return 'closed'
        elif self._file.mode == 'r':
            return 'reading'
        elif self._file.mode == 'r+':
            return 'writing'
        else:
            raise NotImplementedError('Do not understand mode '
                                      f'`{self._file.mode}')
    
    
    def close(self) -> None:
        """ close the currently opened file """
        if self._file is not None:
            self._logger.info(f'Close file `{self.filename}`')
            self._file.close()
            self._file = None
            self._data_length = None  # type: ignore
        
    
    def _create_hdf_dataset(self, name: str, shape: Tuple[int, ...] = tuple()):
        """ create a hdf5 dataset with the given name and data_shape
        
        Args:
            name (str): Identifier of the hdf5 dataset
            shape (tuple): Data shape of the dataset
        """
        if self.compression:
            kwargs = {'chunks': (1,) + shape,
                      'compression': 'gzip'}
        else:
            kwargs = {}
        
        if self._max_length:
            shape = (self._max_length,) + shape
            return self._file.create_dataset(name, shape=shape, dtype=np.double,
                                             **kwargs)
        else:
            return self._file.create_dataset(name, shape=(0,) + shape,
                                             dtype=np.double,
                                             maxshape=(None,) + shape,
                                             **kwargs)
        
                
    def _open(self, mode: str = 'reading', info: InfoDict = None) -> None:
        """ open the hdf file in a particular mode
        
        Args:
            mode (str):
                Determines how the file is opened. Possible values are
                `reading`, `appending`, `writing`, and `closed`.
            info (dict):
                Supplies extra information that is stored in the storage
        """
        import h5py  # lazy loading so it's not a hard dependence
        state = self._file_state

        if mode == 'reading':
            # open file for reading
            if state in ['reading', 'appending', 'writing']:
                return  # we can read data when file is open for writing
            
            # close file to open it again for reading or appending
            if self._file:
                self._file.close()
            self._logger.info(f'Open file `{self.filename}` for reading')
            self._file = h5py.File(self.filename, mode="r")
            self._times = self._file['times']
            self._data = self._file['data']
            for k, v in self._file.attrs.items():
                self.info[k] = json.loads(v)
            if info:
                self.info.update(info)
            
            self._data_shape = self.data.shape[1:]
            self._data_length = self.info.get('data_length')  # type: ignore
            
        elif mode == 'appending':
            # open file for writing without deleting data
            if state in ['appending', 'writing']:
                return  # we are already in a mode where we can append data
            if self.keep_opened and self._is_writing:
                raise RuntimeError('Currently writing data, so mode cannot be '
                                   'switched.')
            if self._file:
                self.close()

            # open file for reading or appending
            self._logger.info(f'Open file `{self.filename}` for appending')
            self._file = h5py.File(self.filename, mode="a")
            
            if 'times' in self._file and 'data' in self._file:
                # extract data from datasets in the existing file
                self._times = self._file['times']
                self._data = self._file['data']
                
                # extract information
                for k, v in self._file.attrs.items():
                    self.info[k] = json.loads(v)
                self._data_shape = self.data.shape[1:]
                self._data_length = \
                                self.info.get('data_length', self.data.shape[0])
                
            else:
                # create new datasets
                self._times = self._create_hdf_dataset("times")
                self._data = self._create_hdf_dataset("data", self.data_shape)
                self._data_length = 0
                    
            if info:
                self.info.update(info)
            
        elif mode == 'writing':
            # open file for writing data; delete potentially present data
            if self._is_writing:
                raise RuntimeError('Currently writing data, so mode cannot be '
                                   'switched.')
            if self._file:
                self.close()
            else:
                ensure_directory_exists(self.filename.parent)
            self._logger.info(f'Open file `{self.filename}` for writing')
            self._file = h5py.File(self.filename, "w")
            self._times = self._create_hdf_dataset("times")
            self._data = self._create_hdf_dataset("data", self.data_shape)
            if info:
                self.info.update(info)
            self._data_length = 0  # start writing from the beginning

        elif mode == 'closed':
            self.close()
            
        else:
            raise NotImplementedError(f'Mode `{mode}` not implemented')


    def __len__(self):
        """ return the number of stored items, i.e., time steps """
        # determine size of data in HDF5 file
        try:
            length = len(self.times)
        except OSError:
            length = 0
            
        if self._data_length is None:
            return length
        else:
            # size of stored data is smaller since preallocation was used
            return min(length, self._data_length)


    @property
    def times(self):
        """ :class:`numpy.ndarray`: The times at which data is available """
        self._open('reading')
        return self._times
        
        
    @property
    def data(self):
        """  :class:`numpy.ndarray`: The actual data for all time """        
        self._open('reading')
        return self._data
    
    
    def clear(self, clear_data_shape: bool = False):
        """ truncate the storage by removing all stored data.
        
        Args:
            clear_data_shape (bool): Flag determining whether the data shape is
                also deleted.
        """
        if self._is_writing:
            self._logger.info('Truncate data in hdf5 file')
            # remove data from opened file
            if 'times' in self._file:
                del self._file['times']
            self._times = self._create_hdf_dataset("times")
                 
            if 'data' in self._file:
                del self._file['data']
            self._data = self._create_hdf_dataset("data", self.data_shape)
            self._data_length = 0  # start writing from start
            
        elif self.filename.is_file():
            self._logger.info('Truncate data by removing hdf5 file')
            self.filename.unlink()
            
        else:
            self._logger.debug('Truncate is no-op since file does not exist')
            
        super().clear(clear_data_shape=clear_data_shape)
            
        
    def start_writing(self, field: FieldBase, info: InfoDict = None) -> None: 
        """ initialize the storage for writing data
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be written to extract the grid
                and the data_shape
            info (dict):
                Supplies extra information that is stored in the storage
        """
        if self._is_writing:
            raise RuntimeError(f'{self.__class__.__name__} is already in '
                               'writing mode')
        
        # delete data if truncation is requested. This is for instance necessary
        # to remove older data with incompatible data_shape
        if self.write_mode == 'truncate' or self.write_mode == 'truncate_once':
            self.clear(clear_data_shape=True)
        
        # initialize the writing, setting current data shape
        super().start_writing(field, info=info)

        # initialize the file for writing with the correct mode        
        self._logger.debug(f'Start writing with mode `{self.write_mode}`')
        if self.write_mode == 'truncate_once':
            self._open('writing', info)
            self.write_mode = 'append'  # do not truncate for next writing
            
        elif self.write_mode == 'truncate':
            self._open('writing', info)

        elif self.write_mode == 'append':        
            self._open('appending', info)

        elif self.write_mode == 'readonly':
            raise RuntimeError('Cannot write in read-only mode')
            
        else:
            raise ValueError(f'Unknown write mode `{self.write_mode}`. '
                             'Possible values are `truncate_once`, '
                             '`truncate`, and `append`')
            
        if not self.keep_opened:
            # store extra information as attributes
            hdf_write_attributes(self._file, self.info)
            
        self._is_writing = True

            
    def append(self, data: np.ndarray, time: Optional[float] = None) -> None:
        """ append a new data set
        
        Args:
            data (:class:`numpy.ndarray`): The actual data
            time (float, optional): The time point associated with the data
        """
        if self.keep_opened:
            if not self._is_writing or self._data_length is None:
                raise RuntimeError('Writing not initialized. Call '
                                   f'`{self.__class__.__name__}.start_writing`')
            
        else:
            # need to reopen the file
            self._open('appending')
        
        # write the new data
        if self._data_length >= len(self._data):
            self._data.resize((self._data_length + 1,) + self.data_shape)
        self._data[self._data_length] = data
        
        # write the new time 
        if time is None:
            if len(self._times) == 0:
                time = 0 
            else:
                time = self._times[self._data_length - 1] + 1
        if self._data_length >= len(self._times):
            self._times.resize((self._data_length + 1,))
        self._times[self._data_length] = time

        self._data_length += 1      
        self.info['data_length'] = self._data_length
        
        if not self.keep_opened:
            self.close()

        
    def end_writing(self) -> None:
        """ finalize the storage after writing.
        
        This makes sure the data is actually written to a file when
        self.keep_opened == False
        """
        if not self._is_writing:
            return  # writing mode was already ended
        self._logger.debug('End writing')
        
        # store extra information as attributes
        hdf_write_attributes(self._file, self.info)
        self._file.flush()
        self.close()
        self._is_writing = False



'''
Integration tests that use multiple modules together

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

from .. import UnitGrid, CartesianGrid, ScalarField, DiffusionPDE, FileStorage
from ..tools.misc import skipUnlessModule



@skipUnlessModule('h5py')
def test_writing_to_storage(tmp_path):
    """ test whether data is written to storage """
    state = ScalarField.random_uniform(UnitGrid([3]))
    pde = DiffusionPDE()
    path = tmp_path / 'test_writing_to_storage.hdf5'
    data = FileStorage(filename=path)
    pde.solve(state, t_range=1.1, dt=0.1, tracker=[data.tracker(0.5)])
    
    assert len(data) == 3



def test_inhomogeneous_bcs():
    """ test simulation with inhomogeneous boundary conditions """
    grid = CartesianGrid([[0, 2*np.pi], [0, 1]], [32, 2],
                         periodic=[True, False])
    state = ScalarField(grid)
    pde = DiffusionPDE(bc=['natural', {'type': 'value', 'value': 'sin(x)'}])
    sol = pde.solve(state, t_range=1e1, dt=1e-2, tracker=None)
    data = sol.get_line_data(extract='project_x')
    np.testing.assert_almost_equal(data['data_y'],
                                   0.9 * np.sin(data['data_x']), decimal=2)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np
from scipy import stats, fftpack

from ..spectral import make_colored_noise
from ...grids import UnitGrid, CartesianGrid



def spectral_density(data, dx=1.):
    """ calculate the power spectral density of a field
    
    Args:
        data (:class:`numpy.ndarray`):
            Data of which the power spectral density will be calculated  
        dx (float or list): The discretizations of the grid either as a single
            number or as an array with a value for each dimension
            
    Returns:
        A tuple with two arrays containing the magnitudes of the wave vectors
        and the associated density, respectively.
    """
    dim = len(data.shape)
    dx = np.broadcast_to(dx, (dim,))
    
    # prepare wave vectors
    k2s = 0
    for i in range(dim):
        k = fftpack.fftfreq(data.shape[i], dx[i])
        k2s = np.add.outer(k2s, k**2)
  
    res = fftpack.fftn(data)

    return np.sqrt(k2s), np.abs(res)**2
    

                            
def test_colored_noise():
    """ test the implementation of the colored noise """
    grid = UnitGrid([64, 64], periodic=True)
    for exponent in [0, -1, 2]:
        scale = np.random.uniform(1, 10)
        noise = make_colored_noise(grid.shape, grid.discretization,
                                   exponent, scale)
        x = noise()
        assert stats.normaltest(x.flat).pvalue > 2e-5, \
               f'Colored noise with exp={exponent} is not normal distributed'
        
        
        
def test_noise_scaling():
    """ compare the noise strength (in terms of the spectral density of
    two different noise sources that should be equivalent) """
    # create a grid
    x, w = 2 + 10 * np.random.random(2)
    size = np.random.randint(128, 256)
    grid = CartesianGrid([[x, x + w]], size, periodic=True)
    
    # colored noise
    noise_colored = make_colored_noise(grid.shape, grid.discretization,
                                       exponent=2)
    
    # divergence of white noise
    shape = (grid.dim, ) + grid.shape
    div = grid.get_operator('divergence', bc='natural')
    
    def noise_div():
        return div(np.random.randn(*shape))
    
    # calculate spectral densities of the two noises
    result = []
    for noise_func in [noise_colored, noise_div]:
        def get_noise():
            k, density = spectral_density(data=noise_func(),
                                          dx=grid.discretization)
            assert k[0] == 0
            assert density[0] == pytest.approx(0)
            return np.log(density[1])  # log of spectral density
        
        # average spectral density of longest length scale
        mean = np.mean([get_noise() for _ in range(64)])
        result.append(mean)
        
    np.testing.assert_allclose(*result, rtol=0.5)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import os
from unittest import mock
import pickle

import numpy as np
import pytest

from .. import trackers
from ...grids import UnitGrid
from ...fields import ScalarField
from ...pdes import DiffusionPDE, CahnHilliardPDE, AllenCahnPDE
from ...solvers import ExplicitSolver, Controller
from ...storage import MemoryStorage
from ...tools.misc import skipUnlessModule, module_available
        


@skipUnlessModule("matplotlib")
def test_plot_tracker(tmp_path):
    """ test whether the plot tracker creates files without errors """
    output_file = tmp_path / "img.png"
    output_folder = tmp_path / "folder/"
    output_folder.mkdir()
    
    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid)
    pde = DiffusionPDE()
    tracker = trackers.PlotTracker(output_file=output_file,
                                   output_folder=output_folder,
                                   interval=0.1, show=False)
    
    pde.solve(state, t_range=0.5, dt=0.005, tracker=tracker, backend='numpy')
    
    assert output_file.exists()
    assert len(list(output_folder.iterdir())) == 6
    
    import matplotlib.pyplot as plt
    plt.close('all')
    
    
    
def test_simple_progress():
    """ simple test for basic progress bar """
    pbar = trackers.ProgressTracker(interval=1)
    field = ScalarField(UnitGrid([3]))
    pbar.initialize(field)
    print(pbar.progress_bar.total)
    pbar.handle(field, 2)
    print(pbar.progress_bar.total)
    pbar.finalize()



def test_progress_no_tqdm(capsys):
    """ test progress bar without tqdm package """
    with mock.patch.dict('sys.modules', {'tqdm': None}):  # @UndefinedVariable
        with pytest.warns(UserWarning):
            test_simple_progress()
    captured = capsys.readouterr()
    assert len(captured.err) > 0



def test_trackers():
    """ test whether simple trackers can be used """
    times = []
    
    def store_time(state, t):
        times.append(t)
        
    def get_data(state):
        return {'integral': state.integral}
    
    devnull = open(os.devnull, 'w')
    data = trackers.DataTracker(get_data, interval=0.1)
    tracker_list = [trackers.PrintTracker(interval=0.1, stream=devnull),
                    trackers.CallbackTracker(store_time, interval=0.1),
                    data]
    if module_available("matplotlib"):
        tracker_list.append(trackers.PlotTracker(interval=0.1, show=False))
    
    grid = UnitGrid([16, 16])
    state = ScalarField.random_uniform(grid, 0.2, 0.3)
    pde = DiffusionPDE()
    pde.solve(state, t_range=1, dt=0.005, tracker=tracker_list)

    devnull.close()
    
    assert times == data.times
    if module_available("pandas"):
        df = data.dataframe
        np.testing.assert_allclose(df['time'], times)
        np.testing.assert_allclose(df['integral'], state.integral)
    if module_available("matplotlib"):
        import matplotlib.pyplot as plt
        plt.close('all')
        


def test_callback_tracker():
    """ test trackers that support a callback """
    data = []
    
    def store_mean_data(state):
        data.append(state.average)
        
    def get_mean_data(state):
        return state.average
    
    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, 0.2, 0.3)
    pde = DiffusionPDE()
    data_tracker = trackers.DataTracker(get_mean_data, interval=0.1)
    tracker_list = [trackers.CallbackTracker(store_mean_data, interval=0.1),
                    data_tracker]
    pde.solve(state, t_range=0.5, dt=0.005, tracker=tracker_list,
              backend='numpy')
    
    np.testing.assert_array_equal(data, data_tracker.data)
   
    data = []
    
    def store_time(state, t):
        data.append(t)
        
    def get_time(state, t):
        return t
    
    grid = UnitGrid([4, 4])
    state = ScalarField.random_uniform(grid, 0.2, 0.3)
    pde = DiffusionPDE()
    data_tracker = trackers.DataTracker(get_time, interval=0.1)
    tracker_list = [trackers.CallbackTracker(store_time, interval=0.1),
                    data_tracker]
    pde.solve(state, t_range=0.5, dt=0.005, tracker=tracker_list,
              backend='numpy')
    
    ts = np.arange(0, 0.55, 0.1)
    np.testing.assert_allclose(data, ts, atol=1e-2)
    np.testing.assert_allclose(data_tracker.data, ts, atol=1e-2)
   
    

def test_data_tracker(tmp_path):
    """ test the DataTracker """
    field = ScalarField(UnitGrid([4, 4]))
    eq = DiffusionPDE()
    
    path = tmp_path / "test_data_tracker.pickle"
    data1 = trackers.DataTracker(lambda f: f.average, filename=path)
    data2 = trackers.DataTracker(lambda f: {'avg': f.average,
                                            'int': f.integral})
    eq.solve(field, 10, tracker=[data1, data2])
    
    with path.open('br') as fp:
        time, data = pickle.load(fp)
    np.testing.assert_allclose(time, np.arange(11))
    assert isinstance(data, list)
    assert len(data) == 11
    
    assert path.stat().st_size > 0
     
    
    
def test_steady_state_tracker():
    """ test the SteadyStateTracker """
    storage = MemoryStorage()
    c0 = ScalarField.random_uniform(UnitGrid([5]))
    pde = DiffusionPDE()
    tracker = trackers.SteadyStateTracker(atol=1e-2, rtol=1e-2)
    pde.solve(c0, 1e3, dt=0.1, tracker=[tracker, storage.tracker(interval=1e2)])
    assert len(storage) < 9  # finished early
    
    

def test_small_tracker_dt():
    """ test the case where the dt of the tracker is smaller than the dt 
    of the simulation """
    storage = MemoryStorage()
    pde = DiffusionPDE()
    c0 = ScalarField.random_uniform(UnitGrid([4, 4]), 0.1, 0.2)
    pde.solve(c0, 1e-2, dt=1e-3, method='explicit',
              tracker=storage.tracker(interval=1e-4))
    assert len(storage) == 11


    
def test_runtime_tracker():
    """ test the RuntimeTracker """
    s = ScalarField.random_uniform(UnitGrid([128]))
    tracker = trackers.RuntimeTracker('0:01')
    sol = ExplicitSolver(DiffusionPDE())
    con = Controller(sol, t_range=1e4, tracker=['progress', tracker])
    con.run(s, dt=1e-3)
    
    
    
def test_consistency_tracker():
    """ test the ConsistencyTracker """
    s = ScalarField.random_uniform(UnitGrid([128]))
    sol = ExplicitSolver(DiffusionPDE(1e3))
    con = Controller(sol, t_range=1e5, tracker=['consistency'])
    with np.errstate(all='ignore'):
        con.run(s, dt=1)
    assert con.info['t_final'] < con.info['t_end']


    
def test_material_conservation_tracker():
    """ test the MaterialConservationTracker """
    state = ScalarField.random_uniform(UnitGrid([8, 8]), 0, 1)
    
    solver = ExplicitSolver(CahnHilliardPDE())
    controller = Controller(solver, t_range=1,
                            tracker=['material_conservation'])
    controller.run(state, dt=1e-3)
    assert controller.info['t_final'] >= 1
    
    solver = ExplicitSolver(AllenCahnPDE())
    controller = Controller(solver, t_range=1,
                            tracker=['material_conservation'])
    controller.run(state, dt=1e-3)
    assert controller.info['t_final'] <= 1
    


"""
Create a movie
==============

This example shows how to create a movie, which is only possible if `ffmpeg` is
installed in a standard location.
"""

from pde import UnitGrid, ScalarField, DiffusionPDE, MemoryStorage, movie_scalar

grid = UnitGrid([16, 16])                           # generate grid
state = ScalarField.random_uniform(grid, 0.2, 0.3)  # generate initial condition

storage = MemoryStorage()               # create storage
tracker = storage.tracker(interval=1)   # create associated tracker

eq = DiffusionPDE()                                 # define the physics
eq.solve(state, t_range=2, dt=0.005, tracker=tracker)

# create movie from stored data
movie_scalar(storage, '/tmp/diffusion.mov')
